/*
 * Command2.c - Command button widget (with arrow)
 */

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Misc.h>
#include <X11/neXtaw/XawInit.h>
#include "Command2P.h"
#include <X11/Xmu/Converters.h>
#include <X11/extensions/shape.h>
#ifdef XPM_TILE
#include <X11/neXtaw/Misc.h>
#endif
#define DEFAULT_HIGHLIGHT_THICKNESS 2
#define DEFAULT_SHAPE_HIGHLIGHT 32767

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

/* Private Data */

static char defaultTranslations[] =
    "<EnterWindow>:	highlight()		\n\
     <LeaveWindow>:	reset()			\n\
     <Btn1Down>:	set()			\n\
     <Btn1Up>:		notify() unset()	";

#define offset(field) XtOffsetOf(Command2Rec, field)
static XtResource resources[] = { 
   {XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer), 
      offset(command.callbacks), XtRCallback, (XtPointer)NULL},
   {XtNhighlightThickness, XtCThickness, XtRDimension, sizeof(Dimension),
      offset(command.highlight_thickness), XtRImmediate,
      (XtPointer) DEFAULT_SHAPE_HIGHLIGHT},
   {XtNshapeStyle, XtCShapeStyle, XtRShapeStyle, sizeof(int),
      offset(command.shape_style), XtRImmediate, (XtPointer)XawShapeRectangle},
   {XtNcornerRoundPercent, XtCCornerRoundPercent, 
	XtRDimension, sizeof(Dimension),
	offset(command.corner_round), XtRImmediate, (XtPointer) 25},
   {XtNshadowWidth, XtCShadowWidth, XtRDimension, sizeof(Dimension),
      offset(threeD.shadow_width), XtRImmediate, (XtPointer) 2},
   {XtNborderWidth, XtCBorderWidth, XtRDimension, sizeof(Dimension),
      XtOffsetOf(RectObjRec,rectangle.border_width), XtRImmediate,
      (XtPointer)0}
};
#undef offset

static Boolean SetValues();
static void Initialize(), Redisplay(), Set(), Reset(), Notify(), Unset();
static void Highlight(), Unhighlight(), Destroy(), PaintCommandWidget();
static void ClassInitialize();
static Boolean ShapeButton();
static void Realize(), Resize();

static XtActionsRec actionsList[] = {
  {"set",		Set},
  {"notify",		Notify},
  {"highlight",		Highlight},
  {"reset",		Reset},
  {"unset",		Unset},
  {"unhighlight",	Unhighlight}
};

#define SuperClass ((LabelWidgetClass)&labelClassRec)

Command2ClassRec command2ClassRec = {
  {
    (WidgetClass) SuperClass,		/* superclass		  */	
    "Command2",				/* class_name		  */
    sizeof(Command2Rec),			/* size			  */
    ClassInitialize,			/* class_initialize	  */
    NULL,				/* class_part_initialize  */
    FALSE,				/* class_inited		  */
    Initialize,				/* initialize		  */
    NULL,				/* initialize_hook	  */
    Realize,				/* realize		  */
    actionsList,			/* actions		  */
    XtNumber(actionsList),		/* num_actions		  */
    resources,				/* resources		  */
    XtNumber(resources),		/* resource_count	  */
    NULLQUARK,				/* xrm_class		  */
    FALSE,				/* compress_motion	  */
    TRUE,				/* compress_exposure	  */
    TRUE,				/* compress_enterleave    */
    FALSE,				/* visible_interest	  */
    Destroy,				/* destroy		  */
    Resize,				/* resize		  */
    Redisplay,				/* expose		  */
    SetValues,				/* set_values		  */
    NULL,				/* set_values_hook	  */
    XtInheritSetValuesAlmost,		/* set_values_almost	  */
    NULL,				/* get_values_hook	  */
    NULL,				/* accept_focus		  */
    XtVersion,				/* version		  */
    NULL,				/* callback_private	  */
    defaultTranslations,		/* tm_table		  */
    XtInheritQueryGeometry,		/* query_geometry	  */
    XtInheritDisplayAccelerator,	/* display_accelerator	  */
    NULL				/* extension		  */
  },  /* CoreClass fields initialization */
  {
    XtInheritChangeSensitive		/* change_sensitive	*/
  },  /* SimpleClass fields initialization */
  {
    XtInheritXaw3dShadowDraw           /* shadowdraw           */
  },  /* ThreeD Class fields initialization */
  {
    0,                                     /* field not used    */
  },  /* LabelClass fields initialization */
  {
    0,                                     /* field not used    */
  },  /* CommandClass fields initialization */
};

  /* for public consumption */
WidgetClass command2WidgetClass = (WidgetClass) &command2ClassRec;

/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static GC 
Get_GC(cbw, fg, bg)
Command2Widget cbw;
Pixel fg, bg;
{
  XGCValues	values;
  
  values.foreground   = fg;
  values.background	= bg;
  values.font		= cbw->label.font->fid;
  values.cap_style = CapProjecting;
  
  if (cbw->command.highlight_thickness > 1 )
    values.line_width   = cbw->command.highlight_thickness;
  else 
    values.line_width   = 0;
  
  if ( cbw->simple.international == True )
      return XtAllocateGC((Widget)cbw, 0, 
		 (GCForeground|GCBackground|GCLineWidth|GCCapStyle),
		 &values, GCFont, 0 );
  else
      return XtGetGC((Widget)cbw,
		 (GCForeground|GCBackground|GCFont|GCLineWidth|GCCapStyle),
		 &values);
}


/* ARGSUSED */
static void 
Initialize(request, new, args, num_args)
Widget request, new;
ArgList args;			/* unused */
Cardinal *num_args;		/* unused */
{
  Command2Widget cbw = (Command2Widget) new;
  int shape_event_base, shape_error_base;

  if (cbw->command.shape_style != XawShapeRectangle
      && !XShapeQueryExtension(XtDisplay(new), &shape_event_base, 
			       &shape_error_base))
      cbw->command.shape_style = XawShapeRectangle;
  if (cbw->command.highlight_thickness == DEFAULT_SHAPE_HIGHLIGHT) {
      if (cbw->command.shape_style != XawShapeRectangle)
	  cbw->command.highlight_thickness = 0;
      else
	  cbw->command.highlight_thickness = DEFAULT_HIGHLIGHT_THICKNESS;
  }
  if (cbw->command.shape_style != XawShapeRectangle) {
    cbw->threeD.shadow_width = 0;
    cbw->core.border_width = 1;
  }
  cbw->command.was_set = False;
      
  cbw->command.normal_GC = Get_GC(cbw, cbw->label.foreground, 
				  cbw->core.background_pixel);
  cbw->command.inverse_GC = Get_GC(cbw, cbw->core.background_pixel, 
				   cbw->label.foreground);
  XtReleaseGC(new, cbw->label.normal_GC);
  cbw->label.normal_GC = cbw->command.normal_GC;

  cbw->command.set = FALSE;
  cbw->command.highlighted = HighlightNone;
}

static Region 
HighlightRegion(cbw)
Command2Widget cbw;
{
  static Region outerRegion = NULL, innerRegion, emptyRegion;
  Dimension s = cbw->threeD.shadow_width;
  XRectangle rect;

  if (cbw->command.highlight_thickness == 0 ||
      cbw->command.highlight_thickness >
      (Dimension) ((Dimension) Min(cbw->core.width, cbw->core.height)/2))
    return(NULL);

  if (outerRegion == NULL) {
    /* save time by allocating scratch regions only once. */
    outerRegion = XCreateRegion();
    innerRegion = XCreateRegion();
    emptyRegion = XCreateRegion();
  }

  rect.x = rect.y = s;
  rect.width = cbw->core.width - 2 * s;
  rect.height = cbw->core.height - 2 * s;
  XUnionRectWithRegion( &rect, emptyRegion, outerRegion );
  rect.x = rect.y += cbw->command.highlight_thickness;
  rect.width -= cbw->command.highlight_thickness * 2;
  rect.height -= cbw->command.highlight_thickness * 2;
  XUnionRectWithRegion( &rect, emptyRegion, innerRegion );
  XSubtractRegion( outerRegion, innerRegion, outerRegion );
  return outerRegion;
}

/***************************
*
*  Action Procedures
*
***************************/

/* ARGSUSED */
static void 
Set(w,event,params,num_params)
Widget w;
XEvent *event;
String *params;		/* unused */
Cardinal *num_params;	/* unused */
{
  Command2Widget cbw = (Command2Widget)w;

  if (cbw->command.set)
    return;

  cbw->command.set= TRUE;
  if (XtIsRealized(w))
    PaintCommandWidget(w, event, (Region) NULL, TRUE);
}

/* ARGSUSED */
static void
Unset(w,event,params,num_params)
Widget w;
XEvent *event;
String *params;		/* unused */
Cardinal *num_params;
{
  Command2Widget cbw = (Command2Widget)w;

  if (!cbw->command.set)
    return;

  cbw->command.set = FALSE;
  if (XtIsRealized(w)) {
    XClearWindow(XtDisplay(w), XtWindow(w));
    PaintCommandWidget(w, event, (Region) NULL, TRUE);
  }
}

/* ARGSUSED */
static void 
Reset(w,event,params,num_params)
Widget w;
XEvent *event;
String *params;		/* unused */
Cardinal *num_params;   /* unused */
{
  Command2Widget cbw = (Command2Widget)w;

  if (cbw->command.set) {
    cbw->command.highlighted = HighlightNone;
    Unset(w, event, params, num_params);
  } else
    Unhighlight(w, event, params, num_params);
}

/* ARGSUSED */
static void 
Highlight(w,event,params,num_params)
Widget w;
XEvent *event;
String *params;		
Cardinal *num_params;	
{
  Command2Widget cbw = (Command2Widget)w;

  if ( *num_params == (Cardinal) 0) 
    cbw->command.highlighted = HighlightWhenUnset;
  else {
    if ( *num_params != (Cardinal) 1) 
      XtWarning("Too many parameters passed to highlight action table.");
    switch (params[0][0]) {
    case 'A':
    case 'a':
      cbw->command.highlighted = HighlightAlways;
      break;
    default:
      cbw->command.highlighted = HighlightWhenUnset;
      break;
    }
  }
  if (XtIsRealized(w))
    PaintCommandWidget(w, event, HighlightRegion(cbw), TRUE);
}

/* ARGSUSED */
static void 
Unhighlight(w,event,params,num_params)
Widget w;
XEvent *event;
String *params;		/* unused */
Cardinal *num_params;	/* unused */
{
  Command2Widget cbw = (Command2Widget)w;

  cbw->command.highlighted = HighlightNone;
  if (XtIsRealized(w))
    PaintCommandWidget(w, event, HighlightRegion(cbw), TRUE);
}

/* ARGSUSED */
static void 
Notify(w,event,params,num_params)
Widget w;
XEvent *event;
String *params;		/* unused */
Cardinal *num_params;	/* unused */
{
  Command2Widget cbw = (Command2Widget)w; 

  /* check to be sure state is still Set so that user can cancel
     the action (e.g. by moving outside the window, in the default
     bindings.
  */
  if (cbw->command.set)
    XtCallCallbackList(w, cbw->command.callbacks, (XtPointer) NULL);
}

/*
 * Repaint the widget window
 */

/************************
*
*  REDISPLAY (DRAW)
*
************************/

/* ARGSUSED */
static void 
Redisplay(w, event, region)
Widget w;
XEvent *event;
Region region;
{
  PaintCommandWidget(w, event, region, FALSE);
}

/*	Function Name: PaintCommandWidget
 *	Description: Paints the command widget.
 *	Arguments: w - the command widget.
 *                 region - region to paint (passed to the superclass).
 *                 change - did it change either set or highlight state?
 *	Returns: none
 */

static void 
PaintCommandWidget(w, event, region, change)
Widget w;
XEvent *event;
Region region;
Boolean change;
{
  Command2Widget cbw = (Command2Widget) w;
  Command2WidgetClass cwclass = (Command2WidgetClass) XtClass (w);
  Boolean very_thick;
  GC norm_gc, rev_gc;
	GC		white = cbw->threeD.top_half_shadow_GC,
			black = cbw->command.normal_GC,
	   	gray = cbw->threeD.bot_shadow_GC;
  Dimension	s = cbw->threeD.shadow_width;
  Dimension XS = cbw->core.width - 20;
  Dimension YS = cbw->core.height / 2;
   
  XSetLineAttributes(XtDisplay(w),black,1,LineSolid,CapProjecting,JoinBevel);
  
  very_thick = cbw->command.highlight_thickness >
               (Dimension)((Dimension) Min(cbw->core.width, cbw->core.height)/2);

/*    
  if (cbw->command.set) {
    cbw->label.normal_GC = cbw->command.inverse_GC;
    XFillRectangle(XtDisplay(w), XtWindow(w), cbw->command.normal_GC,
		   s, s, cbw->core.width - 2 * s, cbw->core.height - 2 * s);
    region = NULL;	*/	/* Force label to repaint text. */
/*  }
  else
    cbw->label.normal_GC = cbw->command.normal_GC;
*/
    XClearArea(XtDisplay(w), XtWindow(w), s, s, cbw->core.width - 2 * s, 
	       cbw->core.height - 2 * s, False);        
    region=NULL;
    if (cbw->label.pixmap == None) {
	if (cbw->command.set) {
	    if (!cbw->command.was_set) {      
		cbw->label.label_x+=s;
		cbw->label.label_y+=s;	
		cbw->command.was_set=True;
	    }
	} else if (cbw->command.was_set) {
	    cbw->label.label_x-=s;
	    cbw->label.label_y-=s;	      
	    cbw->command.was_set=False;
	}
    }
  if (cbw->command.highlight_thickness <= 0) {
    (*SuperClass->core_class.expose) (w, event, region);
    (*cwclass->threeD_class.shadowdraw) (w, event, region, !cbw->command.set);
    return;
  }

/*
 * If we are set then use the same colors as if we are not highlighted. 
 */

  if (cbw->command.set == (cbw->command.highlighted == HighlightNone)) {
    norm_gc = cbw->command.inverse_GC;
    rev_gc = cbw->command.normal_GC;
  } else {
    norm_gc = cbw->command.normal_GC;
    rev_gc = cbw->command.inverse_GC;
  }

  if ( !( (!change && (cbw->command.highlighted == HighlightNone)) ||
	  ((cbw->command.highlighted == HighlightWhenUnset) &&
	   (cbw->command.set))) ) {
    if (very_thick) {
      cbw->label.normal_GC = norm_gc; /* Give the label the right GC. */
      XFillRectangle(XtDisplay(w),XtWindow(w), rev_gc,
		     s, s, cbw->core.width - 2 * s, cbw->core.height - 2 * s);
    }
    else {
      /* wide lines are centered on the path, so indent it */
/*      int offset = cbw->command.highlight_thickness/2; */
	/* Temporarily out
      if (cbw->command.highlight_thickness>0)
	  XDrawRectangle(XtDisplay(w),XtWindow(w), rev_gc, s + offset, s + offset, 
		     cbw->core.width - cbw->command.highlight_thickness - 2 * s,
		     cbw->core.height - cbw->command.highlight_thickness - 2 * s);
	 */
    }
  }
  (*SuperClass->core_class.expose) (w, event, region);
  (*cwclass->threeD_class.shadowdraw) (w, event, region, !cbw->command.set);

	   	
	XDrawLine(XtDisplay(w), XtWindow(w), black, 	XS,   YS, 	XS+4,	YS-4);
	XDrawPoint(XtDisplay(w),XtWindow(w), black, 	XS+5, YS-4);
	XDrawLine(XtDisplay(w), XtWindow(w), black, 	XS+6, YS-2, XS+9,	YS-2);
	XDrawLine(XtDisplay(w), XtWindow(w), black, 	XS+10,YS-2, XS+10,YS-5);
	XDrawLine(XtDisplay(w), XtWindow(w), black, 	XS+11,YS-5, XS+13,YS-5);
	XDrawLine(XtDisplay(w), XtWindow(w), black, 	XS, 	YS, 	XS+4, YS+4);
	
	XDrawLine(XtDisplay(w), XtWindow(w), gray, 	XS+1,	YS, 	XS+4,	YS-3);
	XDrawPoint(XtDisplay(w),XtWindow(w), gray, 	XS+5, YS-3);
	XDrawPoint(XtDisplay(w),XtWindow(w), gray, 	XS+5, YS-2);
	XDrawPoint(XtDisplay(w),XtWindow(w), gray, 	XS+14,YS-5);

	XDrawLine(XtDisplay(w), XtWindow(w), white,	XS+5,	YS+4,	XS+5,	YS+2);
	XDrawLine(XtDisplay(w), XtWindow(w), white,	XS+6,	YS+2,	XS+13,YS+2);
	XDrawLine(XtDisplay(w), XtWindow(w), white,	XS+14,YS+2,	XS+14,YS-4);
}

static void 
Destroy(w)
Widget w;
{
  Command2Widget cbw = (Command2Widget) w;

  /* so Label can release it */
  if (cbw->label.normal_GC == cbw->command.normal_GC)
    XtReleaseGC( w, cbw->command.inverse_GC );
  else
    XtReleaseGC( w, cbw->command.normal_GC );
}

/*
 * Set specified arguments into widget
 */

/* ARGSUSED */
static Boolean 
SetValues (current, request, new, args, num_args)
Widget current, request, new;
ArgList args;
Cardinal *num_args;
{
  Command2Widget oldcbw = (Command2Widget) current;
  Command2Widget cbw = (Command2Widget) new;
  Boolean redisplay = False;

  if ( oldcbw->core.sensitive != cbw->core.sensitive && !cbw->core.sensitive) {
    /* about to become insensitive */
    cbw->command.set = FALSE;
    cbw->command.highlighted = HighlightNone;
    redisplay = TRUE;
  }
  
  if ( (oldcbw->label.foreground != cbw->label.foreground)           ||
       (oldcbw->core.background_pixel != cbw->core.background_pixel) ||
       (oldcbw->command.highlight_thickness != 
                                   cbw->command.highlight_thickness) ||
       (oldcbw->label.font != cbw->label.font) ) 
  {
    if (oldcbw->label.normal_GC == oldcbw->command.normal_GC)
	/* Label has release one of these */
      XtReleaseGC(new, cbw->command.inverse_GC);
    else
      XtReleaseGC(new, cbw->command.normal_GC);

    cbw->command.normal_GC = Get_GC(cbw, cbw->label.foreground, 
				    cbw->core.background_pixel);
    cbw->command.inverse_GC = Get_GC(cbw, cbw->core.background_pixel, 
				     cbw->label.foreground);
    XtReleaseGC(new, cbw->label.normal_GC);
    cbw->label.normal_GC = (cbw->command.set
			    ? cbw->command.inverse_GC
			    : cbw->command.normal_GC);
    
    redisplay = True;
  }

  if ( XtIsRealized(new)
       && oldcbw->command.shape_style != cbw->command.shape_style
       && !ShapeButton(cbw, TRUE))
  {
      cbw->command.shape_style = oldcbw->command.shape_style;
  }

  return (redisplay);
}

static void ClassInitialize()
{    
    XawInitializeWidgetSet();
    XtSetTypeConverter( XtRString, XtRShapeStyle, XmuCvtStringToShapeStyle,
		        (XtConvertArgList)NULL, 0, XtCacheNone, (XtDestructor)NULL );
}


static Boolean
ShapeButton(cbw, checkRectangular)
Command2Widget cbw;
Boolean checkRectangular;
{
    Dimension corner_size = 0;

    if ( (cbw->command.shape_style == XawShapeRoundedRectangle) ) {
	corner_size = (cbw->core.width < cbw->core.height) ? cbw->core.width 
	                                                   : cbw->core.height;
	corner_size = (int) (corner_size * cbw->command.corner_round) / 100;
    }

    if (checkRectangular || cbw->command.shape_style != XawShapeRectangle) {
	if (!XmuReshapeWidget((Widget) cbw, cbw->command.shape_style,
			      corner_size, corner_size)) {
	    cbw->command.shape_style = XawShapeRectangle;
	    return(False);
	}
    }
    return(TRUE);
}

static void Realize(w, valueMask, attributes)
    Widget w;
    Mask *valueMask;
    XSetWindowAttributes *attributes;
{
    (*command2WidgetClass->core_class.superclass->core_class.realize)
	(w, valueMask, attributes);
    ShapeButton( (Command2Widget) w, FALSE);
}

static void Resize(w)
    Widget w;
{
    if (XtIsRealized(w)) 
	ShapeButton( (Command2Widget) w, FALSE);

    (*command2WidgetClass->core_class.superclass->core_class.resize)(w);
}
