/*
 * GroupBox.c - GroupBox widget
 *
 */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xos.h>
#include <X11/neXtaw/XawInit.h>
#include "GroupBoxP.h"
#include <X11/neXtaw/Misc.h>
#include <X11/Xmu/Converters.h>
#include <X11/Xmu/Drawing.h>
#include <stdio.h>
#include <ctype.h>
#ifndef X_NOT_STDC_ENV
#include <stdlib.h>
#else
int abs();
#endif

#define streq(a,b) (strcmp( (a), (b) ) == 0)

#define MULTI_LINE_LABEL 32767

#ifdef CRAY
#define WORD64
#endif

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

/* Private Data */

#define offset(field) XtOffsetOf(GroupBoxRec, field)
static XtResource resources[] = {
    {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel), offset(label.foreground), XtRString, XtDefaultForeground},
    {XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *), offset(label.font),XtRString, XtDefaultFont},
    {XtNfontSet,  XtCFontSet, XtRFontSet, sizeof(XFontSet ), offset(label.fontset),XtRString, XtDefaultFontSet},
    {XtNlabel,  XtCLabel, XtRString, sizeof(String), offset(label.label), XtRString, NULL},
    {XtNencoding, XtCEncoding, XtRUnsignedChar, sizeof(unsigned char), offset(label.encoding), XtRImmediate, (XtPointer)XawTextEncoding8bit},
    {XtNjustify, XtCJustify, XtRJustify, sizeof(XtJustify), offset(label.justify), XtRImmediate, (XtPointer)XtJustifyCenter},

    {XtNgbWidth, XtCWidth, XtRDimension,  sizeof(Dimension), offset(label.gb_width), XtRImmediate, (XtPointer)0},
    {XtNgbHeight,XtCHeight,XtRDimension,  sizeof(Dimension), offset(label.gb_height),XtRImmediate, (XtPointer)0},

    {XtNinternalWidth, XtCWidth, XtRDimension,  sizeof(Dimension), offset(label.internal_width), XtRImmediate, (XtPointer)4},
    {XtNinternalHeight, XtCHeight, XtRDimension, sizeof(Dimension), offset(label.internal_height), XtRImmediate, (XtPointer)2},
    {XtNleftBitmap, XtCLeftBitmap, XtRBitmap, sizeof(Pixmap), offset(label.left_bitmap), XtRImmediate, (XtPointer) None},
    {XtNbitmap, XtCPixmap, XtRBitmap, sizeof(Pixmap), offset(label.pixmap), XtRImmediate, (XtPointer)None},
    {XtNresize, XtCResize, XtRBoolean, sizeof(Boolean), offset(label.resize), XtRImmediate, (XtPointer)True},
    {XtNshadowWidth, XtCShadowWidth, XtRDimension, sizeof(Dimension), offset(threeD.shadow_width), XtRImmediate, (XtPointer) 0},
    {XtNborderWidth, XtCBorderWidth, XtRDimension, sizeof(Dimension), XtOffsetOf(RectObjRec,rectangle.border_width), XtRImmediate, (XtPointer)1},
#ifdef XPM_TILE
	 {XtNbackgroundTile, XtCBackgroundTile, XtRTilePixmap, sizeof(Pixmap), offset(label.background_tile), XtRImmediate, (XtPointer)None}
#endif
};
#undef offset

static void Initialize();
static void Resize();
static void Redisplay();
#ifdef XPM_TILE
static void Realize();
#endif
static Boolean SetValues();
static void ClassInitialize();
static void Destroy();
static XtGeometryResult QueryGeometry();

GroupBoxClassRec groupBoxClassRec = {
  {
    /* superclass	  	*/				(WidgetClass) &threeDClassRec,
    /* class_name	  	*/				"GroupBox",
    /* widget_size	  	*/			sizeof(GroupBoxRec),
    /* class_initialize   	*/		ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited       	*/		FALSE,
    /* initialize	  	*/				Initialize,
    /* initialize_hook		*/		NULL,
#ifdef XPM_TILE
    /* realize		  	*/				Realize,
#else
    /* realize		  	*/				XtInheritRealize,
#endif
    /* actions		  	*/				NULL,
    /* num_actions	  	*/			0,
    /* resources	  	*/				resources,
    /* num_resources	  	*/			XtNumber(resources),
    /* xrm_class	  	*/				NULLQUARK,
    /* compress_motion	  	*/		TRUE,
    /* compress_exposure  	*/		TRUE,
    /* compress_enterleave	*/		TRUE,
    /* visible_interest	  	*/		FALSE,
    /* destroy		  	*/				Destroy,
    /* resize		  	*/				Resize,
    /* expose		  	*/				Redisplay,
    /* set_values	  	*/				SetValues,
    /* set_values_hook		*/		NULL,
    /* set_values_almost	*/		XtInheritSetValuesAlmost,
    /* get_values_hook		*/		NULL,
    /* accept_focus	 	*/			NULL,
    /* version			*/				XtVersion,
    /* callback_private   	*/		NULL,
    /* tm_table		   	*/		NULL,
    /* query_geometry		*/		QueryGeometry,
    /* display_accelerator	*/		XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  {
    /* change_sensitive		*/		XtInheritChangeSensitive
  },
  {
    /* shadowdraw 		*/			XtInheritXaw3dShadowDraw
  },
  {
    /* ignore 			*/				0
  }
};

WidgetClass groupBoxWidgetClass = (WidgetClass)&groupBoxClassRec;


/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static void ClassInitialize()
{
#ifdef XPM_TILE    
    static XtConvertArgRec convertArg[] = {
        {XtWidgetBaseOffset, (XtPointer) XtOffsetOf(WidgetRec, core.screen),
	     sizeof(Screen *)},
        {XtWidgetBaseOffset, (XtPointer) XtOffsetOf(WidgetRec, core.colormap),
	     sizeof(Colormap)}
    };
#endif    
    XawInitializeWidgetSet();
    XtAddConverter( XtRString, XtRJustify, XmuCvtStringToJustify, 
		    (XtConvertArgList)NULL, 0 );
    
#ifdef XPM_TILE
    XtSetTypeConverter(XtRString, XtRTilePixmap, neXtawcvtStringToTilePixmap,
		       convertArg, XtNumber(convertArg),
		       XtCacheByDisplay, (XtDestructor)NULL);
#endif
}

#ifndef WORD64

#define TXT16 XChar2b

#else

#define TXT16 char

static XChar2b *buf2b;
static int buf2blen = 0;

_XawGroupBoxWidth16(fs, str, n)
    XFontStruct *fs;
    char *str;
    int	n;
{
    int i;
    XChar2b *ptr;

    if (n > buf2blen) {
	buf2b = (XChar2b *)XtRealloc((char *)buf2b, n * sizeof(XChar2b));
	buf2blen = n;
    }
    for (ptr = buf2b, i = n; --i >= 0; ptr++) {
	ptr->byte1 = *str++;
	ptr->byte2 = *str++;
    }
    return XTextWidth16(fs, buf2b, n);
}

_XawGroupBoxDraw16(dpy, d, gc, x, y, str, n)
    Display *dpy;
    Drawable d;
    GC gc;
    int x, y;
    char *str;
    int n;
{
    int i;
    XChar2b *ptr;

    if (n > buf2blen) {
	buf2b = (XChar2b *)XtRealloc((char *)buf2b, n * sizeof(XChar2b));
	buf2blen = n;
    }
    for (ptr = buf2b, i = n; --i >= 0; ptr++) {
	ptr->byte1 = *str++;
	ptr->byte2 = *str++;
    }
    XDrawString16(dpy, d, gc, x, y, buf2b, n);
}

#define XTextWidth16 _XawGroupBoxWidth16
#define XDrawString16 _XawGroupBoxDraw16

#endif /* WORD64 */

/*
 * Calculate width and height of displayed text in pixels
 */

static void SetTextWidthAndHeight(lw)
    GroupBoxWidget lw;
{
    XFontStruct	*fs = lw->label.font;
    char *nl;

    if (lw->label.pixmap != None) {
	Window root;
	int x, y;
	unsigned int width, height, bw, depth;
	if (XGetGeometry(XtDisplay(lw), lw->label.pixmap, &root, &x, &y,
			 &width, &height, &bw, &depth)) {
	    lw->label.label_height = height;
	    lw->label.label_width = width;
	    lw->label.label_len = depth;
	    return;
	}
    }
    if ( lw->simple.international == True ) {

      XFontSet	fset = lw->label.fontset;
      XFontSetExtents *ext = XExtentsOfFontSet(fset);

      lw->label.label_height = ext->max_ink_extent.height;
      if (lw->label.label == NULL) {
	  lw->label.label_len = 0;
	  lw->label.label_width = 0;
      }
      else if ((nl = index(lw->label.label, '\n')) != NULL) {
	  char *label;
	  lw->label.label_len = MULTI_LINE_LABEL;
	  lw->label.label_width = 0;
	  for (label = lw->label.label; nl != NULL; nl = index(label, '\n')) {
	      int width = XmbTextEscapement(fset, label, (int)(nl - label));

	      if (width > (int)lw->label.label_width)
		  lw->label.label_width = width;
	      label = nl + 1;
	      if (*label)
		  lw->label.label_height +=
		      ext->max_ink_extent.height;
	  }
	  if (*label) {
	      int width = XmbTextEscapement(fset, label, strlen(label));

	      if (width > (int) lw->label.label_width)
		  lw->label.label_width = width;
	  }
      } else {
	  lw->label.label_len = strlen(lw->label.label);
	  lw->label.label_width =
	      XmbTextEscapement(fset, lw->label.label, (int) lw->label.label_len);
      }

    } else {

        lw->label.label_height = fs->max_bounds.ascent + fs->max_bounds.descent;
        if (lw->label.label == NULL) {
            lw->label.label_len = 0;
            lw->label.label_width = 0;
        }
        else if ((nl = index(lw->label.label, '\n')) != NULL) {
	    char *label;
            lw->label.label_len = MULTI_LINE_LABEL;
            lw->label.label_width = 0;
            for (label = lw->label.label; nl != NULL; nl = index(label, '\n')) {
	        int width;

	        if (lw->label.encoding)
		    width = XTextWidth16(fs, (TXT16*)label, (int)(nl - label)/2);
	        else
		    width = XTextWidth(fs, label, (int)(nl - label));
	        if (width > (int)lw->label.label_width)
		    lw->label.label_width = width;
	        label = nl + 1;
	        if (*label)
		    lw->label.label_height +=
		        fs->max_bounds.ascent + fs->max_bounds.descent;
	    }
	    if (*label) {
	        int width = XTextWidth(fs, label, strlen(label));

	        if (lw->label.encoding)
		    width = XTextWidth16(fs, (TXT16*)label, (int)strlen(label)/2);
	        else
		    width = XTextWidth(fs, label, strlen(label));
	        if (width > (int) lw->label.label_width)
		    lw->label.label_width = width;
	    }
        } else {
	    lw->label.label_len = strlen(lw->label.label);
	    if (lw->label.encoding)
	        lw->label.label_width =
		    XTextWidth16(fs, (TXT16*)lw->label.label,
			         (int) lw->label.label_len/2);
	    else
	        lw->label.label_width =
		    XTextWidth(fs, lw->label.label, (int) lw->label.label_len);
        }

    }
}

#ifdef XPM_TILE
static void Realize(widget, value_mask, attributes)
    Widget widget;
    XtValueMask *value_mask;
    XSetWindowAttributes *attributes;
{
    GroupBoxWidget w = (GroupBoxWidget) widget;

    (*((WidgetClass)&threeDClassRec)->core_class.realize)(widget, value_mask, 
							 attributes);
    if (w->label.background_tile!=None) {
	XSetWindowBackgroundPixmap(XtDisplay(widget), XtWindow(widget),
				   w->label.background_tile);
	XClearWindow(XtDisplay(widget), XtWindow(widget));
    }
}
#endif

static void GetnormalGC(lw)
    GroupBoxWidget lw;
{
    XGCValues	values;

    values.foreground	= lw->label.foreground;
    values.background	= lw->core.background_pixel;
    values.font		= lw->label.font->fid;
    values.graphics_exposures = False;

    lw->label.normal_GC = XtGetGC((Widget)lw,(unsigned) GCForeground | GCBackground | GCFont | GCGraphicsExposures,&values);
}

static void GetgrayGC(lw)
    GroupBoxWidget lw;
{
    XGCValues	values;

    values.foreground = lw->label.foreground;
    values.background = lw->core.background_pixel;
    values.font	      = lw->label.font->fid;
    values.fill_style = FillTiled;
    values.tile       = XmuCreateStippledPixmap(XtScreen((Widget)lw),
						lw->label.foreground, 
						lw->core.background_pixel,
						lw->core.depth);
    values.graphics_exposures = False;

    lw->label.stipple = values.tile;
    lw->label.gray_GC = XtGetGC((Widget)lw,(unsigned) GCForeground | GCBackground | GCFont | GCTile | 
    										GCFillStyle | GCGraphicsExposures, &values);
}

static void compute_bitmap_offsets (lw)
    GroupBoxWidget lw;
{
    /*
     * bitmap will be eventually be displayed at 
     * (internal_width, internal_height + lbm_y)
     */
    if (lw->label.lbm_height != 0) {
	lw->label.lbm_y = (lw->core.height -
			  (lw->threeD.shadow_width * 2 + 
			   lw->label.internal_height * 2 +
			   lw->label.lbm_height)) / 2;
    } else {
	lw->label.lbm_y = 0;
    }
}


static void set_bitmap_info (lw)
    GroupBoxWidget lw;
{
    Window root;
    int x, y;
    unsigned int bw, depth;

    if (!(lw->label.left_bitmap &&
	  XGetGeometry (XtDisplay(lw), lw->label.left_bitmap, &root, &x, &y,
			&lw->label.lbm_width, &lw->label.lbm_height,
			&bw, &depth))) {
	lw->label.lbm_width = lw->label.lbm_height = 0;
    }
    compute_bitmap_offsets (lw);
}



/* ARGSUSED */
static void Initialize(request, new, args, num_args)
    Widget request, new;
    ArgList args;
    Cardinal *num_args;
{
    GroupBoxWidget lw = (GroupBoxWidget) new;

    if (lw->label.label == NULL) 
        lw->label.label = XtNewString(lw->core.name);
    else {
        lw->label.label = XtNewString(lw->label.label);
    }

    GetnormalGC(lw);
    GetgrayGC(lw);

    SetTextWidthAndHeight(lw);

    if (lw->core.height == 0)
        lw->core.height = lw->label.label_height +
			    2 * lw->label.internal_height +
			    2 * lw->threeD.shadow_width;

    set_bitmap_info (lw);		/* need core.height */

    if (lw->core.width == 0)		/* need label.lbm_width */
        lw->core.width = (lw->label.label_width +
			    2 * lw->label.internal_width +
			    2 * lw->threeD.shadow_width +
			    LEFT_OFFSET(lw));

    lw->label.label_x = lw->label.label_y = 0;
    (*XtClass(new)->core_class.resize) ((Widget)lw);

} /* Initialize */

/*
 * Repaint the widget window
 */

/* ARGSUSED */
static void Redisplay(gw, event, region)
    Widget gw;
    XEvent *event;
    Region region;
{
   GroupBoxWidget w = (GroupBoxWidget) gw;
/*
   extern WidgetClass commandWidgetClass;
   GroupBoxWidgetClass lwclass = (GroupBoxWidgetClass) XtClass (gw);
   Position ksy = w->label.label_y; 
*/
   GC gc;

	int len = w->label.label_len;
	char *label = w->label.label;
	Position y = w->label.label_y + w->label.font->max_bounds.ascent;

	Dimension	x1 = 5 ,
	   			x2 = w->label.gb_width - 5,
	   			y1	= 7 ,
	   			y2 = w->label.gb_height - 5,
	   			x12= w->label.label_x - 1,
	   			x13= w->label.label_x + w->label.label_width + 1;
	GC		white = w->threeD.top_half_shadow_GC,
	   	black = w->threeD.bot_shadow_GC;

   gc = XtIsSensitive(gw) ? w->label.normal_GC : w->label.gray_GC;

	XDrawLine(XtDisplay(gw), XtWindow(gw), black, x1, y1, x1, y2);
	XDrawLine(XtDisplay(gw), XtWindow(gw), black, x2, y1, x13, y1);
	XDrawLine(XtDisplay(gw), XtWindow(gw), black, x12, y1, x1, y1);
	XDrawLine(XtDisplay(gw), XtWindow(gw), black, x1+2, y2-1, x2-1, y2-1);
	XDrawLine(XtDisplay(gw), XtWindow(gw), black, x2-1, y2-1, x2-1, y1+2);

	XDrawLine(XtDisplay(gw), XtWindow(gw), white, x1+1, y1+1, x1+1, y2);
	XDrawLine(XtDisplay(gw), XtWindow(gw), white, x2, y1+1, x13, y1+1);
	XDrawLine(XtDisplay(gw), XtWindow(gw), white, x12, y1+1, x1+1, y1+1);
	XDrawLine(XtDisplay(gw), XtWindow(gw), white, x1+1, y2, x2, y2);
	XDrawLine(XtDisplay(gw), XtWindow(gw), white, x2, y2, x2, y1+1);

//	printf("w = %i - h = %i ",w->label.gb_width,w->label.gb_height);
//	puts("debug info");

	if (w->label.encoding) XDrawString16(XtDisplay(gw), XtWindow(gw), gc,w->label.label_x, y, (TXT16*)label, len/2);
	else XDrawString(XtDisplay(gw), XtWindow(gw), gc,w->label.label_x, y, label, len);
}

static void _Reposition(lw, width, height, dx, dy)
    GroupBoxWidget lw;
    Dimension width, height;
    Position *dx, *dy;
{
   Position newPos;
   Position leftedge = LEFT_OFFSET(lw) + lw->threeD.shadow_width;

	switch (lw->label.justify) {

		case XtJustifyLeft   :
	   	newPos = leftedge + 10;
	   	break;

		case XtJustifyRight  :
	   	newPos = width - (lw->label.label_width + lw->label.internal_width + lw->threeD.shadow_width) - 10;
	   	break;

		case XtJustifyCenter :
		default:
	   	newPos = (int)(width - lw->label.label_width) / 2;
	   	break;
   }
	
	if (newPos < (Position)leftedge) newPos = leftedge;
   *dx = newPos - lw->label.label_x;
   lw->label.label_x = newPos;
//   *dy = (newPos = (int)(height - lw->label.label_height) / 2) - lw->label.label_y;
   *dy = (newPos = 0);
	lw->label.label_y = newPos;
   return;
}

static void Resize(w)
    Widget w;
{
    GroupBoxWidget lw = (GroupBoxWidget)w;
    Position dx, dy;

    _Reposition(lw, w->core.width, w->core.height, &dx, &dy);
    compute_bitmap_offsets (lw);
}

/*
 * Set specified arguments into widget
 */

#define PIXMAP 0
#define WIDTH 1
#define HEIGHT 2
#define NUM_CHECKS 3

static Boolean SetValues(current, request, new, args, num_args)
    Widget current, request, new;
    ArgList args;
    Cardinal *num_args;
{
    GroupBoxWidget curlw = (GroupBoxWidget) current;
    GroupBoxWidget reqlw = (GroupBoxWidget) request;
    GroupBoxWidget newlw = (GroupBoxWidget) new;
    int i;
    Boolean was_resized = False, redisplay = False, checks[NUM_CHECKS];

    for (i = 0; i < NUM_CHECKS; i++)
	checks[i] = FALSE;

    for (i = 0; i < *num_args; i++) {
	if (streq(XtNbitmap, args[i].name))
	    checks[PIXMAP] = TRUE;
	if (streq(XtNwidth, args[i].name))
	    checks[WIDTH] = TRUE;
	if (streq(XtNheight, args[i].name))
	    checks[HEIGHT] = TRUE;
    }

    if (newlw->label.label == NULL) {
	newlw->label.label = newlw->core.name;
    }

    /*
     * resize on bitmap change
     */
    if (curlw->label.left_bitmap != newlw->label.left_bitmap) {
	was_resized = True;
    }

    if (curlw->label.encoding != newlw->label.encoding)
	was_resized = True;

    if ( (curlw->label.fontset != newlw->label.fontset) &&
					 curlw->simple.international ){
	was_resized = True;
    }
    if (curlw->label.label != newlw->label.label) {
        if (curlw->label.label != curlw->core.name)
	    XtFree( (char *)curlw->label.label );

	if (newlw->label.label != newlw->core.name) {
	    newlw->label.label = XtNewString( newlw->label.label );
	}
	was_resized = True;
    }

    if (was_resized || (curlw->label.font != newlw->label.font) ||
	(curlw->label.justify != newlw->label.justify) || checks[PIXMAP]) {

	SetTextWidthAndHeight(newlw);
	was_resized = True;
    }

    /* recalculate the window size if something has changed. */
    if (newlw->label.resize && was_resized) {
	if ((curlw->core.height == reqlw->core.height) && !checks[HEIGHT])
	    newlw->core.height = (newlw->label.label_height +
				    2 * newlw->label.internal_height +
				    2 * newlw->threeD.shadow_width);

	set_bitmap_info (newlw);

	if ((curlw->core.width == reqlw->core.width) && !checks[WIDTH])
	    newlw->core.width = (newlw->label.label_width +
				    LEFT_OFFSET(newlw) +
				    2 * newlw->label.internal_width +
				    2 * newlw->threeD.shadow_width);
    }

    if (curlw->label.foreground		!= newlw->label.foreground
	|| curlw->core.background_pixel != newlw->core.background_pixel
	|| curlw->label.font->fid	!= newlw->label.font->fid	  ) {

        /* The Fontset is not in the GC - don't make a new GC if FS changes! */

	XtReleaseGC(new, curlw->label.normal_GC);
	XtReleaseGC(new, curlw->label.gray_GC);
	XmuReleaseStippledPixmap( XtScreen(current), curlw->label.stipple );
	GetnormalGC(newlw);
	GetgrayGC(newlw);
	redisplay = True;
    }

    if ((curlw->label.internal_width != newlw->label.internal_width)
        || (curlw->label.internal_height != newlw->label.internal_height)
	|| was_resized) {
	/* Resize() will be called if geometry changes succeed */
	Position dx, dy;
	_Reposition(newlw, curlw->core.width, curlw->core.height, &dx, &dy);
    }

    return was_resized || redisplay ||
	   XtIsSensitive(current) != XtIsSensitive(new);
}

static void Destroy(w)
    Widget w;
{
    GroupBoxWidget lw = (GroupBoxWidget)w;

    if ( lw->label.label != lw->core.name )
	XtFree( lw->label.label );
    XtReleaseGC( w, lw->label.normal_GC );
    XtReleaseGC( w, lw->label.gray_GC);
    XmuReleaseStippledPixmap( XtScreen(w), lw->label.stipple );
}


static XtGeometryResult QueryGeometry(w, intended, preferred)
    Widget w;
    XtWidgetGeometry *intended, *preferred;
{
    GroupBoxWidget lw = (GroupBoxWidget)w;

    preferred->request_mode = CWWidth | CWHeight;
    preferred->width = (lw->label.label_width + 
			    2 * lw->label.internal_width +
			    2 * lw->threeD.shadow_width +
			    LEFT_OFFSET(lw));
    preferred->height = lw->label.label_height + 
			    2 * lw->label.internal_height +
			    2 * lw->threeD.shadow_width;
    if (  ((intended->request_mode & (CWWidth | CWHeight))
	   	== (CWWidth | CWHeight)) &&
	  intended->width == preferred->width &&
	  intended->height == preferred->height)
	return XtGeometryYes;
    else if (preferred->width == w->core.width &&
	     preferred->height == w->core.height)
	return XtGeometryNo;
    else
	return XtGeometryAlmost;
}
