/*
 *  Recycler - modification of the OffiX waste basket
 *  Copyright (C) 1997  Ocker Kerpershoek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

char progname[]="Recycler";
char title[]="Recycler";
char ConfigPath[]="~/GNUstep/Library/Recycler/";

#include <proplist.h>
#include "headers.h"
#include "file.h"
#include "icon.h"
#include "dialog.h"
#include "windows.h"

struct item {
		Widget	icon;
		Widget	label;
		struct	item	*next;
	};
	
XtAppContext	app;			
Display		*dpy;			
RImage		*recycler,*folder,*file,*tmp;
struct		filenode	*fileroot = NULL;
struct		typenode	*typeroot = NULL;
Pixel 		background,foreground,darkgray=0,lightgray=0;
char		lastone[1000];
Atom 		WM_DELETE_WINDOW, GS_WM_MINIATURIZE_WINDOW, GS_WM_FUNCTION;
struct		item		*itemroot=NULL;
XFontStruct	*defaultfont;
XWMHints 	*hints;
WMAppContext	*WMapp;
WMMenu		*menu;
WMMenu		*info;
WMMenu		*filemenu;
char 		naam[MAX_NAME_LEN];
RContext 	*ctx;
Pixmap 		pix;
char		name[256];
Window 		IconWindow;
int		aantal=-1,realquit=0;

proplist_t 	pl,value,
			DeleteKey,
			PixmapPathKey,
			WastePathKey,
			IconConfigFileKey,
			RootDropKey,
			MaxFilesKey,
			MaxSizeKey;

char		*WastePath,
		*PixmapPath,
		*IconConfigFile,
		*AbsWastePath,
		*AbsConfigPath,
		*AbsIconConfigFile;

long		MaxSize;
int 		MaxFiles,
		Delete,
		RootDrop;
		
Widget	toplevel,			// initial widget
	  Form,				// main form
	    icon,			// the recycler icon
	    border1,			// some borders to make everything look better
	    border2,
	    border3,
	    border4,
	    view,			// viewport with scrollbar
	      form2,			// on this form all icons & labels are placed
	infoShell,			// main info popup shell
	    mainInfoForm,		// main form for the info popup
		infoIcon,
		infoLabel,
     	prefShell,			// main preferences popup shell
	    mainPrefForm,		// main form for the preferences popup
		toggleAlways,		// option indicating that files should always be deleted
		toggleAsk,		// option indicating that the user will be asked if the files should be deleted
		toggleNever,		// option indicating that files should not be deleted
		prefGroupBox1,		// 
		toggleDel,		// option indicating that the user will be asked if the files should be deleted
		toggleNop,		// option indicating that files should not be deleted
		prefGroupBox2,		// 
     	legalShell,			// main popup shell for legal stuff
	    mainLegalForm,		// main form for the legal stuff
		legalText, 		// Text containing GNU-GPL
		legalGroupBox
;

static void DropEventHandler(Widget w,XtPointer,XEvent *,Boolean *);
static void RootDropEventHandler(Widget w,XtPointer,XEvent *,Boolean *);
static void DragEventHandler(Widget w,XtPointer pnt,XEvent *event,Boolean* b);
static void FillIcons();
static void setfont();
static void WMDeleteHandler(Widget,XtPointer,XEvent *,Boolean *);
static void cleanUpIcons(void);
static void readConfig(void);
static void closeConfig(void);
void writeConfig(void);

// menu procedures
static void hide(void *, int , Time );
static void quit(void *, int , Time );
static void ConfirmDelete(void *, int , Time );
static void updateRecycler(void *, int , Time );
static void createmenu(void);


int main(int argc,char *argv[])
{
	int 			initial_x,initial_y;
	RContextAttributes 	attr;
	XGCValues 		gcv;
	int 			mask;
	GC			test;

	readConfig();

	if (ConfigPath[0] == '~') {
		AbsConfigPath = (char*)malloc(sizeof(ConfigPath)+ sizeof(getpwuid(getuid())->pw_dir));
		strcpy(AbsConfigPath,getpwuid(getuid())->pw_dir);
		strcat(AbsConfigPath,ConfigPath+1);
	} else {
		AbsConfigPath = strdup(ConfigPath);
	}
	
	if (WastePath[0] == '~') {
		AbsWastePath = (char*)malloc(sizeof(WastePath)+ sizeof(getpwuid(getuid())->pw_dir));
		strcpy(AbsWastePath,getpwuid(getuid())->pw_dir);
		strcat(AbsWastePath,WastePath+1);
	} else {
		AbsWastePath = strdup(WastePath);
	}

	if (IconConfigFile[0] == '~') {
		AbsIconConfigFile = (char*)malloc(sizeof(IconConfigFile)+ sizeof(getpwuid(getuid())->pw_dir));
		strcpy(AbsIconConfigFile,getpwuid(getuid())->pw_dir);
		strcat(AbsIconConfigFile,IconConfigFile+1);
	} else {
		AbsIconConfigFile = strdup(IconConfigFile);
	}

	mkdir(AbsWastePath,S_IRWXU); // if it didn't exist it will be created

	setDirName(AbsWastePath);
	readDir();
	cleanupDir(MaxFiles,MaxSize);

  	toplevel=XtVaAppInitialize(&app,progname,NULL,0,
		&argc,argv,NULL,
 		XtNtitle,title,
 		XtNresize,(XtArgVal)False,
 		NULL);
				   
	// next 4 lines make the recycler apear in the right bottom of the screen
	initial_y = DisplayHeight(XtDisplay(toplevel),0)-VSIZE-50;
	initial_x = DisplayWidth(XtDisplay(toplevel),0)-HSIZE-90;
	XtVaSetValues(toplevel,XtNx,initial_x,NULL);
	XtVaSetValues(toplevel,XtNy,initial_y,NULL);
	
	DndInitialize(toplevel);
	DndRegisterOtherDrop(DropEventHandler);
	DndRegisterIconDrop(DropEventHandler);
	DndRegisterRootDrop(RootDropEventHandler);
	dpy=XtDisplay(toplevel);


	attr.flags = RC_RenderMode | RC_ColorsPerChannel;
	attr.render_mode = RM_DITHER;
	attr.colors_per_channel = 4;
	ctx = RCreateContext(dpy, DefaultScreen(dpy), &attr);

	GS_WM_MINIATURIZE_WINDOW=XInternAtom(dpy,"_GNUSTEP_WM_MINIATURIZE_WINDOW",False);
	GS_WM_FUNCTION=XInternAtom(dpy,"_GNUSTEP_WM_FUNCTION",False);
	
	sprintf(name,"%sXPM/clear.xpm",AbsConfigPath);
	file = RLoadImage(ctx, name,0);
	if (!file) { puts(RErrorString); exit(1); }
	sprintf(name,"%sXPM/file.xpm",AbsConfigPath);
	tmp = RLoadImage(ctx, name, 0); 
	if (!tmp) { puts(RErrorString); exit(1); }
	RCombineArea(file, tmp, 0, 0, tmp->width, tmp->height, 0, 0);
	RDestroyImage(tmp);

	sprintf(name,"%sXPM/clear.xpm",AbsConfigPath);
	folder = RLoadImage(ctx, name, 0); 
	if (!folder) { puts(RErrorString); exit(1); }
	sprintf(name,"%sXPM/folder.xpm",AbsConfigPath);
	tmp = RLoadImage(ctx, name, 0); 
	if (!tmp) { puts(RErrorString);	exit(1); }
	RCombineArea(folder, tmp, 0, 0, tmp->width, tmp->height, 0, 0);
	RDestroyImage(tmp);

	sprintf(name,"%sXPM/recycler.xpm",AbsConfigPath);
	recycler = RLoadImage(ctx, name, 0); 
	if (!recycler) { puts(RErrorString);	exit(1); }


	DrawWindow();
	InitDialog(toplevel);
	initializeInfoWindow();
	initializeLegalWindow();
	initializePrefWindow();
	setfont();

	readFilesRC();

	FillIcons(); 


	DndRegisterDropWidget(form2, DropEventHandler, NULL);	

	ReadIcons();
	UpdateIcon(aantal==0?0:1);

	XtRealizeWidget(toplevel);


	hints = XAllocWMHints();
	hints->icon_window = IconWindow;
	hints->window_group = XtWindow(toplevel);
	hints->flags = WindowGroupHint | IconWindowHint;
	XSetWMHints(dpy, XtWindow(toplevel), hints);

	createmenu();
		
	{
		Atom prots[2];
		WM_DELETE_WINDOW=XInternAtom(dpy,"WM_DELETE_WINDOW",False);
		prots[0]=WM_DELETE_WINDOW;
		prots[1]=GS_WM_MINIATURIZE_WINDOW;
		XSetWMProtocols(dpy,XtWindow(toplevel),prots,2);
	}
	XtAddEventHandler(toplevel,NoEventMask,True,WMDeleteHandler,NULL);
	WMHideApplication(WMapp);


	gcv.foreground = WhitePixelOfScreen(XtScreen(toplevel)); 
	gcv.background = WhitePixelOfScreen(XtScreen(toplevel));
	mask = GCForeground | GCBackground;
	test = XtGetGC(toplevel,mask,&gcv);

	while (realquit==0) {
		XEvent ev;
		XNextEvent(dpy, &ev);
		WMProcessEvent(WMapp, &ev);
		XtDispatchEvent(&ev);
		XDrawLine(dpy,XtWindow(border1),test,4,0,4,VSIZE-23);
	}

	closeConfig();
	return 0;
}

void setfont(void)
{
	XtVaGetValues(toggleAlways,XtNfont,&defaultfont,NULL);
	XtVaSetValues(prefGroupBox1,XtNfont,defaultfont,NULL);
	XtVaSetValues(prefGroupBox2,XtNfont,defaultfont,NULL);
	XtVaSetValues(infoLabel,XtNfont,defaultfont,NULL);
	XtVaSetValues(legalGroupBox,XtNfont,defaultfont,NULL);
	XtVaSetValues(legalText,XtNfont,defaultfont,NULL);

}

void createmenu(void)
{
	WMapp = WMAppCreateWithMain(dpy,DefaultScreen(dpy),XtWindow(toplevel));
	menu = WMMenuCreate(WMapp,"Recycler");
	info = WMMenuCreate(WMapp,"info...");
	filemenu = WMMenuCreate(WMapp,"file...");
	WMMenuAddSubmenu(menu,"info...",info);
	WMMenuAddSubmenu(menu,"file...",filemenu);
	
	WMMenuAddItem(menu,"Hide",(WMMenuAction)hide,NULL,NULL,NULL);
	WMMenuAddItem(menu,"Quit",(WMMenuAction)quit,NULL,NULL,NULL);
	WMMenuAddItem(info,"Info Panel...",(WMMenuAction)DrawInfoWindow,NULL,NULL,NULL);
	WMMenuAddItem(info,"Legal...",(WMMenuAction)DrawLegalWindow,NULL,NULL,NULL);
	WMMenuAddItem(info,"Preferences...",(WMMenuAction)DrawPrefWindow,NULL,NULL,NULL);
	WMMenuAddItem(filemenu,"Delete all",(WMMenuAction)ConfirmDelete,NULL,NULL,NULL);
	WMMenuAddItem(filemenu,"update",(WMMenuAction)updateRecycler,NULL,NULL,NULL);
	
	WMAppSetMainMenu(WMapp,menu);
	WMRealizeMenus(WMapp);
	WMAppAddWindow(WMapp, XtWindow(toplevel));   
}

void updateRecycler(void *a, int b, Time c)
{
	readDir();  
	FillIcons(); 
	UpdateIcon(aantal==0?0:1);
}

void hide(void *a, int b, Time c)
{
	WMHideApplication(WMapp);
}

void quit(void *a, int b, Time c)
{
	exit(0);
}

int find(char *s,char *w,int hoeveelste)
{
	int found,j,i = 0, nummer = 0;

	if (strlen(s) < strlen(w)) return (-1);
	while(i <= (int)(strlen(s)-strlen(w)))	{
		found = 1;
		for (j=0;j < (int)strlen(w);j++) if (s[i+j] != w[j]) found = 0;
		if (found == 1) nummer++;
		if ((found == 1) && (nummer == hoeveelste)) {
			return i;
		}
		i++;
	}
	return (-1);
}

void writeConfig(void)
{
	char	tmpstr[80];
	proplist_t valueNew;
	
	sprintf(tmpstr,"%i",Delete);
	valueNew = PLMakeString(tmpstr);
	pl = PLInsertDictionaryEntry(pl,DeleteKey,valueNew);

	sprintf(tmpstr,"%i",RootDrop);
	valueNew = PLMakeString(tmpstr);
	pl = PLInsertDictionaryEntry(pl,RootDropKey,valueNew);

	sprintf(tmpstr,"%i",MaxFiles);
	valueNew = PLMakeString(tmpstr);
	pl = PLInsertDictionaryEntry(pl,MaxFilesKey,valueNew);

	sprintf(tmpstr,"%li",MaxSize);
	valueNew = PLMakeString(tmpstr);
	pl = PLInsertDictionaryEntry(pl,MaxSizeKey,valueNew);

	PLSave(pl,NO);
}

void closeConfig(void)
{
	writeConfig();
	
	PLRelease(DeleteKey);
	PLRelease(PixmapPathKey);
	PLRelease(WastePathKey);
	PLRelease(IconConfigFileKey);
	PLRelease(RootDropKey);
	PLRelease(MaxFilesKey);
	PLRelease(MaxSizeKey);
	PLRelease(pl);
}

void readConfig(void)
{
	char	regel[1024];
	
	strcpy(regel,ConfigPath);
	strcat(regel,"recycler.pl");
		
	if ((pl = PLGetProplistWithPath(regel))==NULL) {
		puts("Recycler : error reading configuration file.");
		puts("Does ~/GNUstep/Library/Recycler/recycler.pl exist ?");
		exit(0);
	}

	DeleteKey		= PLMakeString("Delete");
	PixmapPathKey		= PLMakeString("PixmapPath");
	WastePathKey		= PLMakeString("WastePath");
	IconConfigFileKey	= PLMakeString("IconConfigFile");
	RootDropKey		= PLMakeString("RootDrop");
	MaxFilesKey		= PLMakeString("MaxFiles");
	MaxSizeKey		= PLMakeString("MaxSize"); 

	value = PLGetDictionaryEntry(pl,DeleteKey);
	Delete = atoi(PLGetString(value));

	value = PLGetDictionaryEntry(pl,RootDropKey);
	RootDrop = atoi(PLGetString(value));
	
	value = PLGetDictionaryEntry(pl,MaxFilesKey);
	MaxFiles = atoi(PLGetString(value));

	value = PLGetDictionaryEntry(pl,MaxSizeKey);
	MaxSize = atol(PLGetString(value));

	value = PLGetDictionaryEntry(pl,PixmapPathKey);
	PixmapPath = strdup(PLGetString(value));

	value = PLGetDictionaryEntry(pl,WastePathKey);
	WastePath = strdup(PLGetString(value));

	value = PLGetDictionaryEntry(pl,IconConfigFileKey);
	IconConfigFile = strdup(PLGetString(value));

	writeConfig();	
}

// this function handles window manager events like minimize and close
// both minimize and close hide the aplication.

void WMDeleteHandler(Widget widget,XtPointer a,XEvent *event,Boolean *b)
{
	if(event->type != ClientMessage)	return;
	if(event->xclient.format != 32)		return;
	if (((Atom)event->xclient.data.l[0] == GS_WM_MINIATURIZE_WINDOW) || ((Atom)event->xclient.data.l[0] == WM_DELETE_WINDOW)) {
		WMHideApplication(WMapp);
		return;
	}
}

void cleanUpIcons(void)
{
	struct	item	*n=itemroot,*m;

	if (itemroot != NULL) {
		while (n->next != NULL) {
			m = n;
			n = n->next;
			if (m->icon != NULL) XtDestroyWidget(m->icon);
			if (m->label != NULL) XtDestroyWidget(m->label);
			free(m);
		}
		if (n->icon != NULL) XtDestroyWidget(n->icon);
		if (n->label != NULL) XtDestroyWidget(n->label);
		free(n);
	}
	itemroot = NULL;
}

// (re)draw all the deleted items
// all files in the trash directory get an icon

void FillIcons()
{
	struct filenode *x = fileroot;
	struct item	*n,*m;
	int 		count=0,hdist=25;

	cleanUpIcons();
	
	while (x != NULL) {
	
		n = (struct item *) malloc(sizeof(struct item));
		n->next = NULL;

		if (x->dir == 0) RConvertImage(ctx, pixmapForType(x->naam), &pix);
		else RConvertImage(ctx, folder, &pix);

		n->icon = XtVaCreateManagedWidget("Icon",
			toggleWidgetClass, form2,
			XtNleft,XtChainLeft,
			XtNright,XtChainLeft,
			XtNtop,XtChainTop,
			XtNbottom,XtChainTop,
			XtNbackground,background,
			XtNwidth,(XtArgVal)50,
			XtNheight,(XtArgVal)50,
			XtNhorizDistance,hdist,
			XtNvertDistance,(XtArgVal)5,
			XtNbitmap,pix,
			XtNlabel,x->naam,NULL); 
		DndRegisterDragWidget(n->icon, DragEventHandler, NULL);
		n->label = XtVaCreateManagedWidget("label",
			labelWidgetClass, form2,
			XtNleft,XtChainLeft,
			XtNright,XtChainLeft,
			XtNtop,XtChainTop,
			XtNbottom,XtChainTop,
			XtNbackground,background,
			XtNwidth,100,
			XtNhorizDistance,(XtArgVal)hdist-21,
			XtNvertDistance,(XtArgVal)55,
			XtNborderWidth,(XtArgVal)0,
			XtNfont,defaultfont,
			XtNlabel,x->naam,
			NULL); 
		count++;
		hdist += 100;
		x = x->next;

		if (itemroot == NULL) itemroot = n;
		else {
			m = itemroot;
			while (m->next != NULL) m = m->next;
			m->next = n;
		}
	}
	if (aantal!=-1)	UpdateIcon(count==0?0:1);
	aantal = count;
}



// this function does not ask for confirmation but simply deletes everything
// the function is activated by clicking on the recycler icon when the recycler is
// maximized.

void ConfirmDelete(void *a, int b, Time c)
{
	struct	filenode 	*f = fileroot;
	char 	command[1000];

	if (Delete==3) return;
	if (Delete==2) {
		if (DrawDialogWindow("delete all files?","No","Yes",defaultfont)!=DIALOG_BUTTON2) return;
	}
	
	while (f != NULL) {
		strcpy(command,"rm -R ");
		strcat(command,naam);
		strcat(command,f->naam);
		strcat(command," ");
		system(command); 
		f = f->next;
	}

	readDir();  
	FillIcons(); 
	UpdateIcon(aantal==0?0:1);		// very probably 0
}



// this function copies the droped file(s) or directory to the trash dir and
// afterwards deletes the files. This goes wrong if you drag a file from the 
// recycler to the recycler (pretty useless too), the file is then first copied 
// and then deleted resulting in permenently destoying the file.

void DropEventHandler(Widget w,XtPointer a,XEvent *event,Boolean *b)
{
	int Type=DndDataType(event);
	char command[1000];
	unsigned char *Data;
	unsigned long DataSize;
 
	if(Type==DndNotDnd) return;
	
	if(DndSourceWindow(event)==XtWindow(w)) {
		puts("drop from myself");
		return;
	}
	
	DndGetData(&Data,&DataSize);
	
	if(Type==DndFile || Type==DndExe || Type==DndDir) {	

		strcpy(command,"cp -r \"");
		strcat(command,(char*)Data);
		strcat(command,"\" ");
		strcat(command,strcat(getpwuid(getuid())->pw_dir,"/.trash/"));
		strcat(command," ; rm -R \"");
		strcat(command,(char*)Data);
		strcat(command,"\"");
		system(command);
		readDir();  
		cleanupDir(MaxFiles,MaxSize);
		FillIcons(); 
		return;
	}
	if(Type==DndFiles) {
		while(*Data) {
			strcpy(command,"cp -r \"");
			strcat(command,(char*)Data);
			strcat(command,"\" ");
			strcat(command,strcat(getpwuid(getuid())->pw_dir,"/.trash/"));
			strcat(command,"; rm -R \"");
			strcat(command,(char*)Data);
			strcat(command,"\"");
			system(command);
			Data+=strlen((char*)Data)+1;
		}
		readDir();  
		cleanupDir(MaxFiles,MaxSize);
		FillIcons(); 
	}

}

void RootDropEventHandler(Widget w,XtPointer a,XEvent *event,Boolean *b)
{
	int Type=DndDataType(event);
	char command[1000];
	unsigned char *Data;
	unsigned long DataSize;
	
	if (Type == DndNotDnd) return;
	if (RootDrop == 2) return;
	if (DndSourceWindow(event) == XtWindow(w)) {
		puts("drop from myself");
		return;
	}
	
	DndGetData(&Data,&DataSize);

	if (Delete==3) return;
	if (Delete==2) {
		if(Type==DndFiles) strcpy(command,"Delete selected files ?");
		else strcpy(command,"Delete selected file ?");
		if (DrawDialogWindow(command,"No","Yes",defaultfont)!=DIALOG_BUTTON2) return;
	}
	
	if(Type==DndFile || Type==DndExe || Type==DndDir) {	
		strcpy(command,"rm -R \"");
		strcat(command,(char*)Data);
		strcat(command,"\"");
		system(command);
		readDir();  
		cleanupDir(MaxFiles,MaxSize);
		FillIcons(); 
		return;
	}
	if(Type==DndFiles) {
		while(*Data) {
			strcpy(command,"rm -R \"");
			strcat(command,(char*)Data);
			strcat(command,"\"");
			system(command);
			Data+=strlen((char*)Data)+1;
		}
		readDir();  
		cleanupDir(MaxFiles,MaxSize);
		FillIcons(); 
	}

}



// this function copies the name and path of the draged item in a data-structure
// that is passed to Dnd.

void DragEventHandler(Widget w,XtPointer a,XEvent *event,Boolean *b)
{
	int DataType=DndFile;
	unsigned long DataSize=0;
	unsigned char *Data;
	char *widgetnaam;
	struct filenode *n;

	Data = (unsigned char *) malloc(100);
	widgetnaam = (char *) malloc(100);

	XtVaGetValues(w,XtNlabel,&widgetnaam,NULL);

	if (strcmp(lastone,widgetnaam)!=0) {
		strcpy(lastone,widgetnaam);
		n = fileroot;	
		while((n != NULL) && (strcmp(n->naam,widgetnaam)!=0)) n = n->next;
		if (n != NULL) {
			if (n->dir == 0) DataType = DndFile;
			else DataType = DndDir;
			strcpy((char *)Data,naam);
			strcat((char *)Data,n->naam);
			DataSize = (unsigned long)(strlen((char *)Data)+1);
		}
		DndSetData(DataType,Data,DataSize);
	}
	DndHandleDragging(w,event);
}
